/*
 * Copyright 2019 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * You may not use this file except in compliance with the terms and conditions 
 * set forth in the accompanying LICENSE.TXT file.
 *
 * THESE MATERIALS ARE PROVIDED ON AN "AS IS" BASIS. AMAZON SPECIFICALLY DISCLAIMS, WITH 
 * RESPECT TO THESE MATERIALS, ALL WARRANTIES, EXPRESS, IMPLIED, OR STATUTORY, INCLUDING 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.
*/

// This skill sample demonstrates sending directives and receiving events from an Echo connected gadget from an Alexa skill.
// This skill uses the Alexa Skills Kit SDK (v2). Please visit https://alexa.design/cookbook for additional
// examples on implementing slots, dialog management, session persistence, api calls, and more.

const Alexa = require('ask-sdk-core');
const Util = require('./util');
const Common = require('./common');

// https://developer.amazon.com/docs/echo-button-skills/keep-session-open.html
// https://stackoverflow.com/questions/43443602/alexa-skills-set-sdk-increase-timeout-of-skill
// The audio tag to include background music
const BG_MUSIC = '<audio src="soundbank://soundlibrary/ui/gameshow/amzn_ui_sfx_gameshow_countdown_loop_64s_minimal_01"/>';
// const BG_MUSIC = '';

// The namespace of the custom directive to be sent by this skill
const NAMESPACE = 'Custom.Mindstorms.Gadget';

// The name of the custom directive to be sent this skill
const NAME_CONTROL = 'control';

const EventHandlerDuration = 64150; // An event handler can only be active for a maximum duration of 90 seconds.

const LaunchRequestHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'LaunchRequest';
    },
    handle: async function(handlerInput) {

        const request = handlerInput.requestEnvelope;
        const {apiEndpoint, apiAccessToken} = request.context.System;
        const apiResponse = await Util.getConnectedEndpoints(apiEndpoint, apiAccessToken);
        if ((apiResponse.endpoints || []).length === 0) {
            return handlerInput.responseBuilder
            .speak(`I couldn't find an EV3 Brick connected to this Echo device. Please check to make sure your EV3 Brick is connected, and try again.`)
//            .withShouldEndSession(true)     // null/false/true  http://ask-sdk-node-typedoc.s3-website-us-east-1.amazonaws.com/classes/responsebuilder.html
            .getResponse();
        }

        // Store the gadget endpointId to be used in this skill session
        const endpointId = apiResponse.endpoints[0].endpointId || [];
        Util.putSessionAttribute(handlerInput, 'endpointId', endpointId);

        // Set skill duration to 10 minutes (ten 60-seconds interval)
        Util.putSessionAttribute(handlerInput, 'duration', 15);

        // Set the token to track the event handler
        const token = handlerInput.requestEnvelope.request.requestId;
        Util.putSessionAttribute(handlerInput, 'token', token);

        // Construct the directive
        const directive = Util.build(endpointId, NAMESPACE, NAME_CONTROL,
            {
                type: 'new session'
            });

        let speechOutput = "Welcome, the robot will be active for fifteen minutes. You can start issuing commands. I will play music. It is needed to keep session alive. Start your commands with Alexa wake word.";
        return handlerInput.responseBuilder
            .speak(speechOutput + BG_MUSIC)
//            .withShouldEndSession(null)     // null/false/true  http://ask-sdk-node-typedoc.s3-website-us-east-1.amazonaws.com/classes/responsebuilder.html
//            .reprompt("Awaiting commands" + BG_MUSIC)
            .addDirective(Util.buildStartEventHandler(token, EventHandlerDuration, {}))
            .addDirective(directive)
            .getResponse();
    }
};

// Add the speed value to the session attribute.
// This allows other intent handler to use the specified speed value
// without asking the user for input.
const SetSpeedIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && Alexa.getIntentName(handlerInput.requestEnvelope) === 'SetSpeedIntent';
    },
    handle: function (handlerInput) {

        // Bound speed to (1-100)
        let speed = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Speed');
        speed = Math.max(1, Math.min(100, parseInt(speed)));
        Util.putSessionAttribute(handlerInput, 'speed', speed);

        let speechOutput = `speed set to ${speed} percent.`;
        return handlerInput.responseBuilder
            .speak(speechOutput + BG_MUSIC)
//            .reprompt("Awaiting commands" + BG_MUSIC)
            .getResponse();
    }
};

// Construct and send a custom directive to the connected gadget with
// data from the MoveIntent request.
const moves = ["follow line", "forward", "backward", "stop", "brake"];
const MoveIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && (Alexa.getIntentName(handlerInput.requestEnvelope) === 'MoveIntent'
                || Alexa.getIntentName(handlerInput.requestEnvelope) === 'AMAZON.StopIntent');
    },
    handle: function (handlerInput) {
        // Get data from session attribute
        const attributesManager = handlerInput.attributesManager;

        var move, distance, condition, sensor, speed;

        if (Alexa.getIntentName(handlerInput.requestEnvelope) === 'AMAZON.StopIntent') {
            move = "stop";
            speed = "0";
        }
        else {
            move = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Move');
            speed = attributesManager.getSessionAttributes().speed || "65";
        }

        if (move === "forwards" || move === "straight")
            move = "forward";
        else if (move === "backwards" || move === "back")
            move = "backward";

        if (moves.includes(move)) {
            distance = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Distance') || "0";
            condition = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Condition') || "";
            sensor = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Sensor') || "";

            if (condition === "cube")
                condition = "obstacle";

            const endpointId = attributesManager.getSessionAttributes().endpointId || [];
    
            // Construct the directive with the payload containing the move parameters
            const directive = Util.build(endpointId, NAMESPACE, NAME_CONTROL,
                {
                    type: 'move',
                    move: move,
                    distance: distance,
                    condition: condition,
                    sensor: sensor,
                    speed: speed
                });
    
            const speechOutput = (move === "follow line" || move === "stop" || move === "brake")
                ? `${move}`
                : `Move ${move}`;
    
            return handlerInput.responseBuilder
                .speak(speechOutput + BG_MUSIC)
//                .reprompt("Awaiting commands" + BG_MUSIC)
                .addDirective(directive)
                .getResponse();
        }
        else
            return handlerInput.responseBuilder
                .speak("Say it again" + BG_MUSIC)
//                .reprompt("Awaiting commands" + BG_MUSIC)
                .getResponse();
    }
};

// Construct and send a custom directive to the connected gadget with
// data from the TurnIntent request.
const directions = ["right", "left", "around"];
const TurnIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && Alexa.getIntentName(handlerInput.requestEnvelope) === 'TurnIntent';
    },
    handle: function (handlerInput) {
        const request = handlerInput.requestEnvelope;
        var direction = Alexa.getSlotValue(request, 'Direction');

        if (direction === "clockwise")
            direction = "right";
        else if (direction === "counterclockwise")
            direction = "left";
        else if (direction === "over")
            direction = "around";

        if (directions.includes(direction)) {
            // Angle is optional, use default "90" degrees if not available
            var angle = Alexa.getSlotValue(request, 'Angle') || "0";

            const speechOutput = (direction === "around" || direction === "over" || angle === "0")
                ? `Turn ${direction}`
                : `Turn ${direction} ${angle} degrees`;

            // Send this to the robot instead "Turn around"
            if (direction === "around") {
                direction = "left";
                angle = "180";
            }

            // Get data from session attribute
            const attributesManager = handlerInput.attributesManager;
            const speed = attributesManager.getSessionAttributes().speed || "40";
            const endpointId = attributesManager.getSessionAttributes().endpointId || [];

            // Construct the directive with the payload containing the turn parameters
            const directive = Util.build(endpointId, NAMESPACE, NAME_CONTROL,
                {
                    type: 'turn',
                    direction: direction,
                    angle: angle,
                    speed: speed
                });

            return handlerInput.responseBuilder
                .speak(speechOutput + BG_MUSIC)
//                .reprompt("Awaiting commands" + BG_MUSIC)
                .addDirective(directive)
                .getResponse();
        }
        else
            return handlerInput.responseBuilder
                .speak("Say it again" + BG_MUSIC)
//                .reprompt("Awaiting commands" + BG_MUSIC)
                .getResponse();
    }
};

// Handle the SessionExtendIntent request.
const SessionExtendIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && Alexa.getIntentName(handlerInput.requestEnvelope) === 'SessionExtendIntent';
    },
    handle: function (handlerInput) {
        return handlerInput.responseBuilder
            .speak("Ok" + BG_MUSIC)
//            .reprompt("Awaiting commands" + BG_MUSIC)
            .getResponse();
    }
};

// Construct and send a custom directive to the connected gadget with data from
// the CommandIntent.
const commands = ["enable", "disable", "check", "change mode", "grab", "drop", "good job"];
const CommandIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && Alexa.getIntentName(handlerInput.requestEnvelope) === 'CommandIntent';
    },
    handle: function (handlerInput) {

        let command = Alexa.getSlotValue(handlerInput.requestEnvelope, 'Command');
        if (command) {

            if (command === "take")
                command = "grab";
            else if (command === "leave")
                command = "drop";
            
            if (commands.includes(command)) {
                const attributesManager = handlerInput.attributesManager;
                let endpointId = attributesManager.getSessionAttributes().endpointId || [];
        
                // Construct the directive with the payload containing the command parameters
                let directive = Util.build(endpointId, NAMESPACE, NAME_CONTROL,
                    {
                        type: 'command',
                        command: command
                    });
        
                let speechOutput;
                
                if(command === "good job")
                    speechOutput = `Thank you `
                    + `<audio src="soundbank://soundlibrary/gameshow/gameshow_01"/> `
                    + `<audio src="soundbank://soundlibrary/ui/gameshow/amzn_ui_sfx_gameshow_intro_01"/> `
                    + `<audio src="soundbank://soundlibrary/ui/gameshow/amzn_ui_sfx_gameshow_countdown_loop_32s_full_01"/> `
                    + BG_MUSIC;
                else
                    speechOutput = `OK` + BG_MUSIC;
                
                return handlerInput.responseBuilder
                    .speak(speechOutput)
        //            .reprompt("Awaiting commands" + BG_MUSIC)            
                    .addDirective(directive)
                    .getResponse();
            }
        }

        return handlerInput.responseBuilder
            .speak("Can you repeat that?")
            .withShouldEndSession(false)
            .getResponse();
    }
};

const EventsReceivedRequestHandler = {
    // Checks for a valid token and endpoint.
    canHandle(handlerInput) {
        let { request } = handlerInput.requestEnvelope;
        console.log('Request type: ' + Alexa.getRequestType(handlerInput.requestEnvelope));
        if (request.type !== 'CustomInterfaceController.EventsReceived') return false;

        const attributesManager = handlerInput.attributesManager;
        let sessionAttributes = attributesManager.getSessionAttributes();
        let customEvent = request.events[0];

        // Validate event token
        if (sessionAttributes.token !== request.token) {
            console.log("Event token doesn't match. Ignoring this event");
            return false;
        }

        // Validate endpoint
        let requestEndpoint = customEvent.endpoint.endpointId;
        if (requestEndpoint !== sessionAttributes.endpointId) {
            console.log("Event endpoint id doesn't match. Ignoring this event");
            return false;
        }
        return true;
    },
    handle(handlerInput) {

        console.log("== Received Custom Event ==");
        let customEvent = handlerInput.requestEnvelope.request.events[0];
        let payload = customEvent.payload;
        let name = customEvent.header.name;

        let speechOutput;
        if (name === 'White line') {
            speechOutput = "White line";
        } else if (name === 'Detected') {
            speechOutput = "Obstacle detected";
        } else if (name === 'Enabled') {
            speechOutput = "Gyro Sensor is enabled.";
        } else if (name === 'Disabled') {
            speechOutput = "Gyro Sensor is disabled.";
        } else if (name === 'Gyro sensor OK') {
            speechOutput = "Gyro sensor works fine.";
        } else if (name === 'Drift') {
            speechOutput = "The robot has detected drift in Gyro sensor. This is a known problem. While LEGO engineers work hard to create a permanent solution, I would suggest a workaround. You may issue a change mode command to try it.";
        } else if (name === 'Exception') {
            speechOutput = "Gyro Sensor caused exception. It has been disabled.";
        } else if (name === 'Speech') {
            speechOutput = payload.speechOut;
        } else {
            speechOutput = "Event not recognized. Awaiting new command.";
        }
        return handlerInput.responseBuilder
            .speak(speechOutput + BG_MUSIC, "REPLACE_ALL")
//            .reprompt("Awaiting commands" + BG_MUSIC)
            .getResponse();
    }
};

const ExpiredRequestHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'CustomInterfaceController.Expired'
    },
    handle(handlerInput) {
        console.log("== Custom Event Expiration Input ==");

        // Set the token to track the event handler
        const token = handlerInput.requestEnvelope.request.requestId;
        Util.putSessionAttribute(handlerInput, 'token', token);

        const attributesManager = handlerInput.attributesManager;
        let duration = attributesManager.getSessionAttributes().duration || 0;
        if (duration > 0) {
            Util.putSessionAttribute(handlerInput, 'duration', --duration);

            // Extends skill session
            const speechOutput = `${duration} minutes remaining.`;
            return handlerInput.responseBuilder
                .addDirective(Util.buildStartEventHandler(token, EventHandlerDuration, {}))
                .speak(BG_MUSIC)
                .getResponse();
        }
        else {
            // End skill session
            return handlerInput.responseBuilder
                .speak("Skill duration expired. Goodbye.")
                .withShouldEndSession(true)
                .getResponse();
        }
    }
};

const ExitIntentHandler = {
    canHandle(handlerInput) {
        return Alexa.getRequestType(handlerInput.requestEnvelope) === 'IntentRequest'
            && (Alexa.getIntentName(handlerInput.requestEnvelope) === 'ExitIntent'
                || Alexa.getIntentName(handlerInput.requestEnvelope) === 'AMAZON.CancelIntent');
    },
    handle(handlerInput) {
        // Get data from session attribute
        const attributesManager = handlerInput.attributesManager;
        const endpointId = attributesManager.getSessionAttributes().endpointId || []; 

        // Set skill duration to 0
        Util.putSessionAttribute(handlerInput, 'duration', 0);

        // Construct the directive
        const directive = Util.build(endpointId, NAMESPACE, NAME_CONTROL,
            {
                type: 'session ended'
            });

        const speakOutput = 'Goodbye!';
        return handlerInput.responseBuilder
            .speak(speakOutput)
            .withShouldEndSession(true)
            .addDirective(directive)
            .getResponse();
    }
};

// The SkillBuilder acts as the entry point for your skill, routing all request and response
// payloads to the handlers above. Make sure any new handlers or interceptors you've
// defined are included below. The order matters - they're processed top to bottom.
exports.handler = Alexa.SkillBuilders.custom()
    .addRequestHandlers(
        LaunchRequestHandler,
        SetSpeedIntentHandler,
        CommandIntentHandler,
        MoveIntentHandler,
        TurnIntentHandler,
        SessionExtendIntentHandler,
        EventsReceivedRequestHandler,
        ExpiredRequestHandler,
        Common.HelpIntentHandler,
        ExitIntentHandler, // Common.CancelAndStopIntentHandler,
        Common.SessionEndedRequestHandler,
        Common.IntentReflectorHandler, // make sure IntentReflectorHandler is last so it doesn't override your custom intent handlers
    )
    .addRequestInterceptors(Common.RequestInterceptor)
    .addErrorHandlers(
        Common.ErrorHandler,
    )
    .lambda();
